/*
 *  Licensed Materials - Property of IBM
 *  5725-G92 (C) Copyright IBM Corp. 2011, 2013. All Rights Reserved.
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
*/

package com.ibm.impact.demo.services;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.ServletConfig;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;

import org.apache.wink.common.model.multipart.BufferedInMultiPart;
import org.apache.wink.common.model.multipart.InPart;
import org.apache.wink.json4j.JSONArray;
import org.apache.wink.json4j.JSONException;
import org.apache.wink.json4j.JSONObject;

/**
 * 
 * MultiFileUploadDataService is written to be specifically compatible with the dojox.form.FileUploader Dojo widget.  More
 * specifically, it is written to be compatible with the Dojo client included with this sample that makes use of the
 * dojox.form.FileUploader Dojo widget.  The method processing the multipart/form-data submission (upload) is generic; it is not
 * specific to the Dojo widget.  Other methods, such as getImage and deleteImages are specific to the Dojo client sample.
 *
 */
@Path("upload")
public class MultiFileUploadDataService {
	
	// The root of the filesystem area in which to store uploaded files:
	private static String rootPath;
	
	// Some regular expression patterns to help us parse the part headers of the
	// individual parts of the multipart/form-data payload sent to the POST
	
	// Semicolon between 0 or more whitespace chars
	private static final Pattern SEMICOLON_PATTERN = Pattern.compile("\\s*+;\\s*+");  
	// All non-whitespace chars between 0 or more whitespace chars
	private static final Pattern NON_WHITESPACE_PATTERN = Pattern.compile("\\s*+(\\S*)\\s*+");
	// All non-quote chars between exactly one double-quote char
	private static final Pattern BETWEEN_QUOTES_PATTERN = Pattern.compile("\"([^\"]*)\"");  
	
	/**
	 * upload method receives a JAX-RS Wink BufferedInMultiPart payload.  Iterate over the "parts" of this inbound multipart/form-data
	 * payload to retrieve each part of the message body.  Messages with a Content-Type of multipart/form-data are typically submissions
	 * of HTML forms with standard form input fields plus file input fields.
	 * 
	 * The response from this method is specifically compatible with the provided Dojo client sample.  Your client code may
	 * require some other type of response.
	 * 
	 * Also, REST best practice says we should return a "Location" header with the URL of the newly created resource (this is a POST,
	 * after all).  Unfortunately, this resource is not "pure" REST; we may create multiple files as a result of the data submitted
	 * to this POST method.  So, for the sake of consistency, we do not return a "Location" header for this resource.
	 * 
	 * Note that if an exception occurs it will propagate up to the web container, and the web container decides what to do with it.
	 * Some will return a 500 Internal Server Error with the stack trace in the message body, which is not always desirable.  If this
	 * is the case in your situation, surround the method with a try/catch(Throwable) and throw new WebApplicationException with a 
	 * more terse message.
	 * 
	 * @param servletConfig injected by the JAX-RS engine
	 * @param httpHeaders injected by the JAX-RS engine
	 * @param uriInfo injected by the JAX-RS engine
	 * @param bimp the HTTP messsage payload, already parsed into "parts"
	 * @return JSONArray.   By leveraging the produces annotation, the JSONArray will be serialized in JSON and returned to the client.
	 * @throws IOException 
	 * @throws Exception
	 */
	@POST
    @Consumes(MediaType.MULTIPART_FORM_DATA)
    @Produces(MediaType.APPLICATION_JSON)
    public JSONArray upload(
    		@Context ServletConfig servletConfig, 
    		@Context HttpHeaders httpHeaders, 
    		@Context UriInfo uriInfo, 
    		BufferedInMultiPart bimp) throws IOException  {

		// Read the servlet configuration to determine the location in the file system
		// where uploaded files will be written.
    	initRootPath(servletConfig);
    	
        JSONArray jsonArray = new JSONArray();
        // Access the list of parts from the multi-part HTTP request.
        List<InPart> parts = bimp.getParts();

        // Iterate through the parts of the multi-part HTTP request.
        for (int i = 0; i < parts.size(); i++) {
        	// Look for Content-Disposition header denoting a part of interest.
            if (parts.get(i).getHeaders().containsKey("Content-Disposition")) {
            	// Found a worthy part.  Read the Content-Disposition header to access the file information.
                ArrayList<String> list = (ArrayList<String>)parts.get(i).getHeaders().get("Content-Disposition");
                
                // Parse the Content-Disposition header, getting a map of the important portions.
                // There's only one Content-Disposition header per part, so it's safe to use list.get(0):
                Map<String, String> cdHeaderMap = parseContentDispositionHeader(list.get(0));
                
                // Look for the filename in the map of header information.
                String filename = cdHeaderMap.get("filename");
                if (filename != null) {  // this Part is a file
                	// This part is a file.  Create a representative file in the configured directory
                    byte[] bytes = readInputStream(parts.get(i).getInputStream());
                    String fullPathToFile = rootPath + "/" + filename;
                    File file = new File(fullPathToFile);

                    // Create the file if it doesn't already exist.
                    if (!file.exists()) {
                        file.createNewFile();
                    } // TODO: consider an else clause here to protect against overwriting existing file
                    
                    // Write that data into the file on the local file system.
                    FileOutputStream fos = new FileOutputStream(file);
                    fos.write(bytes);
                    fos.close();

                    // Build up the return string as required by our sample Dojo client
                    try {
	                    JSONObject jsonObject = new JSONObject();
	                    jsonObject.put("uri", getFileURL(uriInfo, filename));
	                    jsonObject.put("name", filename);
	                    jsonArray.add(jsonObject);
                    } catch (JSONException e) {
                    	throw new WebApplicationException(Response.Status.BAD_REQUEST);                    	
                    }
                }
            }
        }
        return jsonArray;

    }

    /**
     * getImage returns the raw FileInputStream of the requested file.  Smarter clients (Firefox, Chrome, etc) do not
     * require the Content-Type header to determine the returned file type.  For demonstration, we do set the outbound
     * Content-Type header in this method.  Consider that the filename extension does not always match the actual
     * file type.  This is often how web browsers start their check for file types, so we do it here.
     * 
     * @param filename the name of the file being requested
     * @return raw FileInputStream output via Response object
     * @throws FileNotFoundException
     */
	@GET
    @Path("{filename}")
    public Response getImage(@PathParam("filename") String filename)
    throws FileNotFoundException {
		File file = new File(rootPath + "/" + filename);
		if (file.exists()) {
			String fileNameExtension = file.toString().substring(0, file.toString().lastIndexOf('.'));

			MediaType mediaType = MediaType.APPLICATION_OCTET_STREAM_TYPE;  // this default is standard for browsers
			if (fileNameExtension.equals("jpg")) {
				mediaType = new MediaType("image", "jpeg");
			} else if (fileNameExtension.equals("gif")) {
				mediaType = new MediaType("image", "gif");
			} else if (fileNameExtension.equals("png")) {
				mediaType = new MediaType("image", "png");
			} else if (fileNameExtension.equals("zip")) {
				mediaType = new MediaType("image", "zip");
			}

			return Response.ok().type(mediaType).entity(new FileInputStream(file)).build();
		}
		// requested resource does not exist; return a 404
    	throw new WebApplicationException(Response.Status.NOT_FOUND);
    }
	
	/**
	 * In the spirit of REST, we support DELETE on the uploaded resource(s).
	 * 
	 * Calling this method will delete the previously uploaded file.
	 * 
	 * @param filename passed to us on the URL path
	 * @return
	 */
	@DELETE
	@Path("{filename}")
	public Response deleteImage(@PathParam("filename") String filename) {
		File file = new File(rootPath + "/" + filename);
		if (file.exists()) {
			file.delete();
		} else {
			// can't delete a resource that doesn't exist!  Return a 404
			throw new WebApplicationException(Response.Status.NOT_FOUND);
		}
		return Response.ok().build();
	}

    /**
     * If the just-uploaded file is stored in a web-accessible place, we can return a URL
     * that points directly to it.  If it's not in a web-accessible place, we have to
     * return a URL that goes through our GET method, which will retrieve the file from
     * disk and feed back the raw FileInputStream data with the appropriate Content-Type
     * header.
     * 
     * In this sample, we just feed back the raw file data every time, so the URL we give back
     * here points to the URL that hits our getImage method.
     * 
     * @param filename
     * @return relative URL for the uploaded file
     */
    private String getFileURL(UriInfo uriInfo, String filename) {
    	return uriInfo.getRequestUri().toString() + "/" + filename;
	}

    /**
     * This is a utility method used to read the data from a multi-part form data 
     * input stream into a byte array which represents the contents of a file being uploaded.
     * 
     * @param is the InputStream to be read
     * @return byte array of the read InputStream
     * @throws IOException
     */
    private byte[] readInputStream(InputStream is) throws IOException {
        byte[] bytes = new byte[is.available()];
        int b;
        int i = 0;
        while ((b = is.read()) != -1) {
            bytes[i] = (byte)b;
            ++i;
        }
        is.close();
        return bytes;
    }
    
    /**
     * Later versions of JAX-RS Wink have built-in capability to parse the Content-Disposition headers in the
     * "parts".  Since we don't know what version of Wink on which you'll be running this sample, we have to
     * parse it here.
     * 
     * @param value the Content-Disposition header value
     * @return Map of key/value pairs of attributes in the Content-Disposition header, including the Content-Disposition
     *   attribute itself.
     * @throws IllegalArgumentException
     */
    private Map<String, String> parseContentDispositionHeader(String value) {
        String[] all = SEMICOLON_PATTERN.split(value);
        int i = 0;
        Matcher m = (NON_WHITESPACE_PATTERN).matcher(all[i]);

        // Create a map into which valuable portions of the Content-Distribution
        // header will be stored.
        Map<String, String> mapToReturn = new HashMap<String, String>();
        
        // Extract the headerType value and store it in the return map.
        String headerType = "";
        if (m.find()) {
            headerType = m.group(1);
        }
        mapToReturn.put("Content-Disposition", headerType);
        
        // Iterate the list of items in the Content-Disposition header.
        // Extract the "name" and "filename" items from the list and store them in
        // the map to return.
        for(i = 1; i < all.length; i++) {
            if (all[i].startsWith("name")) {
                m = (BETWEEN_QUOTES_PATTERN).matcher(all[i]);
                if (m.find()) {
                	mapToReturn.put("name", m.group(1));
                }
            } else if (all[i].startsWith("filename")) {
                m = (BETWEEN_QUOTES_PATTERN).matcher(all[i]);
                if (m.find()) {
                	mapToReturn.put("filename", m.group(1));
                }
            }
        }
        return mapToReturn;
    }
    
    /**
     * This is a utility method called up early during the execution of this class' resource
     * method in order to identify the directory in which files should be written.  The 
     * "javax.servlet.context.tempdir" property is searched for in the servlet configuration.
     * If the property is not found, a WebApplicationException will be thrown.
     * 
     * @param servletConfig the already-injected ServletConfig object
     */
	private static synchronized void initRootPath(ServletConfig servletConfig) {
		// No need to rerun the logic here if the rootPath was already determined.
		if (rootPath == null) {
			// Access the required property in the servlet configuration.
			//rootPath = (File)servletConfig.getServletContext().getAttribute("javax.servlet.context.tempdir");
			rootPath = servletConfig.getServletContext().getRealPath("/") + File.separator + "images" + File.separator;
			// Ensure the required property was found.
			if ((rootPath == null) || (rootPath.equals(""))) {
				throw new WebApplicationException(Response.status(Response.Status.INTERNAL_SERVER_ERROR).entity("No tempdir specified for application.").build());
			}
		}
		System.out.println("rootPath -->" + rootPath);
	}

    
}
